/* Based on https://github.com/ESultanik/mtwister (public domain)
 *
 * Original notice:
 * An implementation of the MT19937 Algorithm for the Mersenne Twister
 * by Evan Sultanik.  Based upon the pseudocode in: M. Matsumoto and
 * T. Nishimura, "Mersenne Twister: A 623-dimensionally
 * equidistributed uniform pseudorandom number generator," ACM
 * Transactions on Modeling and Computer Simulation Vol. 8, No. 1,
 * January pp.3-30 1998.
 *
 * http://www.sultanik.com/Mersenne_twister
 */

#define UPPER_MASK		0x80000000
#define LOWER_MASK		0x7fffffff
#define TEMPERING_MASK_B	0x9d2c5680
#define TEMPERING_MASK_C	0xefc60000

#include "mersenne_twister.h"

#define STATE_VECTOR_M      397 /* changes to MT_STATE_VECTOR_LENGTH also require changes to this */

void mt_rand_init(struct mt_rand *inst, uint32_t seed)
{
	/* set initial seeds to mt[MT_STATE_VECTOR_LENGTH] using the generator
	 * from Line 25 of Table 1 in: Donald Knuth, "The Art of Computer
	 * Programming," Vol. 2 (2nd Ed.) pp.102.
	 */
	inst->mt[0] = seed & 0xffffffff;
	for(inst->index=1; inst->index<MT_STATE_VECTOR_LENGTH; inst->index++) {
		inst->mt[inst->index] = (6069 * inst->mt[inst->index-1]) & 0xffffffff;
	}
}

uint32_t mt_rand_get_u32(struct mt_rand *inst)
{
	uint32_t y;
	static uint32_t mag[2] = {0x0, 0x9908b0df}; /* mag[x] = x * 0x9908b0df for x = 0,1 */
	if(inst->index >= MT_STATE_VECTOR_LENGTH || inst->index < 0) {
		/* generate MT_STATE_VECTOR_LENGTH words at a time */
		int32_t kk;
		if(inst->index >= MT_STATE_VECTOR_LENGTH+1 || inst->index < 0) {
			mt_rand_init(inst, 4357);
		}
		for(kk=0; kk<MT_STATE_VECTOR_LENGTH-STATE_VECTOR_M; kk++) {
			y = (inst->mt[kk] & UPPER_MASK) | (inst->mt[kk+1] & LOWER_MASK);
			inst->mt[kk] = inst->mt[kk+STATE_VECTOR_M] ^ (y >> 1) ^ mag[y & 0x1];
		}
		for(; kk<MT_STATE_VECTOR_LENGTH-1; kk++) {
			y = (inst->mt[kk] & UPPER_MASK) | (inst->mt[kk+1] & LOWER_MASK);
			inst->mt[kk] = inst->mt[kk+(STATE_VECTOR_M-MT_STATE_VECTOR_LENGTH)] ^ (y >> 1) ^ mag[y & 0x1];
		}
		y = (inst->mt[MT_STATE_VECTOR_LENGTH-1] & UPPER_MASK) | (inst->mt[0] & LOWER_MASK);
		inst->mt[MT_STATE_VECTOR_LENGTH-1] = inst->mt[STATE_VECTOR_M-1] ^ (y >> 1) ^ mag[y & 0x1];
		inst->index = 0;
	}
	y = inst->mt[inst->index++];
	y ^= (y >> 11);
	y ^= (y << 7) & TEMPERING_MASK_B;
	y ^= (y << 15) & TEMPERING_MASK_C;
	y ^= (y >> 18);
	return y;
}
